#include "Eigen/Dense"
#include <iostream>
#include <cassert>

template<typename T, int M, int N>
void Regression_NE(const Eigen::Matrix<T,M,N> &A,
                   Eigen::Matrix<T,N,1> &p,
                   const Eigen::Matrix<T,M,1> &y)
{
  std::cout << "NE" << std::endl;
  p = (A.transpose()*A).llt().solve(A.transpose()*y);
}

template<typename T, int M, int N>
void Regression_QR(const Eigen::Matrix<T,M,N> &A,
                   Eigen::Matrix<T,N,1> &p,
                   const Eigen::Matrix<T,M,1> &y)
{
  std::cout << "QR" << std::endl;
  p = A.colPivHouseholderQr().solve(y);
}

template<typename T, int M, int N>
void Regression_SVD(const Eigen::Matrix<T,M,N> &A,
                   Eigen::Matrix<T,N,1> &p,
                   const Eigen::Matrix<T,M,1> &y)
{
  std::cout << "SVD" << std::endl;
  p = A.bdcSvd(Eigen::ComputeFullU | Eigen::ComputeFullV).solve(y);
}
int main(int argc, char* argv[]) {
  assert(argc==3); int m=std::stoi(argv[1]), n=std::stoi(argv[2]);

  using T=double;
  using MT=Eigen::Matrix<T,Eigen::Dynamic,Eigen::Dynamic>;
  using VT=Eigen::Matrix<T,Eigen::Dynamic,1>;

  MT A=MT::Random(m,n); VT p(n), y=VT::Random(m);

  Regression_NE(A,p,y);

  std::cout << "p= " << p.transpose() << std::endl;
  std::cout << "||A*p-y|| = " << (A*p-y).squaredNorm() << std::endl;
  std::cout << "||A^T*A*p-A^T*y||=" << (A.transpose()*A*p-A.transpose()*y).norm() << std::endl;

  Regression_QR(A,p,y);

  std::cout << "p= " << p.transpose() << std::endl;
  std::cout << "||A*p-y|| = " << (A*p-y).squaredNorm() << std::endl;
  std::cout << "||A^T*A*p-A^T*y||=" << (A.transpose()*A*p-A.transpose()*y).norm() << std::endl;

  Regression_SVD(A,p,y);

  std::cout << "p= " << p.transpose() << std::endl;
  std::cout << "||A*p-y|| = " << (A*p-y).squaredNorm() << std::endl;
  std::cout << "||A^T*A*p-A^T*y||=" << (A.transpose()*A*p-A.transpose()*y).norm() << std::endl;

}
